#!/usr/bin/env python

"""
Interpretation and preparation of iMIP content, together with a content handling
mechanism employed by specific recipients.
"""

from datetime import date, datetime
from email.mime.text import MIMEText
from pytz import timezone, UnknownTimeZoneError
from vCalendar import parse, ParseError, to_dict
import imip_store
import re

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

# iCalendar date and datetime parsing (from DateSupport in MoinSupport).

date_icalendar_regexp_str = ur'(?P<year>[0-9]{4})(?P<month>[0-9]{2})(?P<day>[0-9]{2})'
datetime_icalendar_regexp_str = date_icalendar_regexp_str + \
    ur'(?:' \
    ur'T(?P<hour>[0-2][0-9])(?P<minute>[0-5][0-9])(?P<second>[0-6][0-9])' \
    ur'(?P<utc>Z)?' \
    ur')?'

match_date_icalendar = re.compile(date_icalendar_regexp_str, re.UNICODE).match
match_datetime_icalendar = re.compile(datetime_icalendar_regexp_str, re.UNICODE).match

# Content interpretation.

def get_items(d, name, all=True):

    """
    Get all items from 'd' with the given 'name', returning single items if
    'all' is specified and set to a false value and if only one value is
    present for the name. Return None if no items are found for the name.
    """

    if d.has_key(name):
        values = d[name]
        if not all and len(values) == 1:
            return values[0]
        else:
            return values
    else:
        return None

def get_item(d, name):
    return get_items(d, name, False)

def get_value_map(d, name):

    """
    Return a dictionary for all items in 'd' having the given 'name'. The
    dictionary will map values for the name to any attributes or qualifiers
    that may have been present.
    """

    items = get_items(d, name)
    if items:
        return dict(items)
    else:
        return {}

def get_values(d, name, all=True):
    if d.has_key(name):
        values = d[name]
        if not all and len(values) == 1:
            return values[0][0]
        else:
            return map(lambda x: x[0], values)
    else:
        return None

def get_value(d, name):
    return get_values(d, name, False)

def get_utc_datetime(d, name):
    value, attr = get_item(d, name)
    dt = get_datetime(value, attr)
    return to_utc_datetime(dt)

def to_utc_datetime(dt):
    if not dt:
        return None
    elif isinstance(dt, datetime):
        return dt.astimezone(timezone("UTC"))
    else:
        return dt

def format_datetime(dt):
    if not dt:
        return None
    elif isinstance(dt, datetime):
        return dt.strftime("%Y%m%dT%H%M%SZ")
    else:
        return dt.strftime("%Y%m%d")

def get_address(value):
    return value.startswith("mailto:") and value[7:] or value

def get_uri(value):
    return value.startswith("mailto:") and value or "mailto:%s" % value

def get_datetime(value, attr):
    try:
        tz = attr.has_key("TZID") and timezone(attr["TZID"]) or None
    except UnknownTimeZoneError:
        tz = None

    if attr.get("VALUE") in (None, "DATE-TIME"):
        m = match_datetime_icalendar(value)
        if m:
            dt = datetime(
                int(m.group("year")), int(m.group("month")), int(m.group("day")),
                int(m.group("hour")), int(m.group("minute")), int(m.group("second"))
                )

            # Impose the indicated timezone.
            # NOTE: This needs an ambiguity policy for DST changes.

            tz = m.group("utc") and timezone("UTC") or tz or None
            if tz is not None:
                return tz.localize(dt)
            else:
                return dt

    if attr.get("VALUE") == "DATE":
        m = match_date_icalendar(value)
        if m:
            return date(
                int(m.group("year")), int(m.group("month")), int(m.group("day"))
                )
    return None

# Handler mechanism objects.

def handle_itip_part(part, recipients, handlers):

    """
    Handle the given iTIP 'part' for the given 'recipients' using the given
    'handlers'. Return a list of responses, each response being a tuple of the
    form (is-outgoing, message-part).
    """

    method = part.get_param("method")

    # Decode the data and parse it.

    f = StringIO(part.get_payload(decode=True))

    itip = parse_object(f, part.get_content_charset(), "VCALENDAR")

    # Ignore the part if not a calendar object.

    if not itip:
        return []

    # Require consistency between declared and employed methods.

    if get_value(itip, "METHOD") == method:

        # Look for different kinds of sections.

        all_results = []

        for name, cls in handlers:
            for details in get_values(itip, name) or []:

                # Dispatch to a handler and obtain any response.

                handler = cls(details, recipients)
                result = methods[method](handler)()

                # Concatenate responses for a single calendar object.

                if result:
                    response_method, part = result
                    outgoing = method != response_method
                    all_results.append((outgoing, part))

        return all_results

    return []

def parse_object(f, encoding, objtype):

    """
    Parse the iTIP content from 'f' having the given 'encoding'. Return None if
    the content was not readable or suitable.
    """

    try:
        try:
            doctype, attrs, elements = obj = parse(f, encoding=encoding)
            if doctype == objtype:
                return to_dict(obj)[objtype][0]
        finally:
            f.close()
    except (ParseError, ValueError):
        pass

    return None

def to_part(method, calendar):

    """
    Write using the given 'method', the 'calendar' details to a MIME
    text/calendar part.
    """

    encoding = "utf-8"
    out = StringIO()
    try:
        imip_store.to_stream(out, imip_store.make_calendar(calendar, method), encoding)
        part = MIMEText(out.getvalue(), "calendar", encoding)
        part.set_param("method", method)
        return part

    finally:
        out.close()

class Handler:

    "General handler support."

    def __init__(self, details, recipients):

        """
        Initialise the handler with the 'details' of a calendar object and the
        'recipients' of the object.
        """

        self.details = details
        self.recipients = set(recipients)

        self.uid = get_value(details, "UID")
        self.sequence = get_value(details, "SEQUENCE")
        self.dtstamp = get_value(details, "DTSTAMP")

        self.store = imip_store.FileStore()

        try:
            self.publisher = imip_store.FilePublisher()
        except OSError:
            self.publisher = None

    def get_items(self, name, all=True):
        return get_items(self.details, name, all)

    def get_item(self, name):
        return get_item(self.details, name)

    def get_value_map(self, name):
        return get_value_map(self.details, name)

    def get_values(self, name, all=True):
        return get_values(self.details, name, all)

    def get_value(self, name):
        return get_value(self.details, name)

    def get_utc_datetime(self, name):
        return get_utc_datetime(self.details, name)

    def filter_by_recipients(self, values):
        return self.recipients.intersection(map(get_address, values))

    def require_organiser_and_attendees(self):

        """
        Return the organiser and attendees for the current object, filtered by
        the recipients of interest. Return None if no identities are eligible.
        """

        attendee_map = self.get_value_map("ATTENDEE")
        organiser = self.get_item("ORGANIZER")

        # Only provide details for recipients who are also attendees.

        attendees = {}
        for attendee in map(get_uri, self.filter_by_recipients(attendee_map)):
            attendees[attendee] = attendee_map[attendee]

        if not attendees and not organiser:
            return None

        return organiser, attendees

    def have_new_object(self, attendee, objtype):

        """
        Return whether the current object is new to the 'attendee' for the
        given 'objtype'.
        """

        f = self.store.get_event(attendee, self.uid)
        event = f and parse_object(f, "utf-8", objtype)

        # If found, compare SEQUENCE and potentially DTSTAMP.

        if event:
            sequence = get_value(event, "SEQUENCE")
            dtstamp = get_value(event, "DTSTAMP")

            # If the request refers to an older version of the event, ignore
            # it.

            old_dtstamp = self.dtstamp < dtstamp

            if sequence is not None and (
                int(self.sequence) < int(sequence) or
                int(self.sequence) == int(sequence) and old_dtstamp
                ) or old_dtstamp:

                return False

        return True

# Handler registry.

methods = {
    "ADD"            : lambda handler: handler.add,
    "CANCEL"         : lambda handler: handler.cancel,
    "COUNTER"        : lambda handler: handler.counter,
    "DECLINECOUNTER" : lambda handler: handler.declinecounter,
    "PUBLISH"        : lambda handler: handler.publish,
    "REFRESH"        : lambda handler: handler.refresh,
    "REPLY"          : lambda handler: handler.reply,
    "REQUEST"        : lambda handler: handler.request,
    }

# vim: tabstop=4 expandtab shiftwidth=4
